@@ -10,6 +10,7 @@
 #ifdef RKTIO_SYSTEM_UNIX
 # include <unistd.h>
 # include <netinet/in.h>
+# include <netinet/tcp.h>
 # include <netdb.h>
 # include <sys/socket.h>
 # include <sys/types.h>
@@ -99,31 +100,31 @@ typedef struct SOCKADDR_IN rktio_unspec_address;
 
 #endif
 
-static void do_get_socket_error(rktio_t *rktio) {
-  rktio->errid = SOCK_ERRNO();
+static void do_get_socket_error(rktio_err_t *err) {
+  err->errid = SOCK_ERRNO();
 #ifdef RKTIO_SYSTEM_WINDOWS
-  rktio->errkind = RKTIO_ERROR_KIND_WINDOWS;
+  err->errkind = RKTIO_ERROR_KIND_WINDOWS;
 #else
-  rktio->errkind = RKTIO_ERROR_KIND_POSIX;
+  err->errkind = RKTIO_ERROR_KIND_POSIX;
 #endif
 }
-#define get_socket_error() do_get_socket_error(rktio)
+#define get_socket_error() do_get_socket_error(&rktio->err)
 
-static void do_set_socket_error(rktio_t *rktio, int errid) {
-  rktio->errid = errid;
+static void do_set_socket_error(rktio_err_t *err, int errid) {
+  err->errid = errid;
 #ifdef RKTIO_SYSTEM_WINDOWS
-  rktio->errkind = RKTIO_ERROR_KIND_WINDOWS;
+  err->errkind = RKTIO_ERROR_KIND_WINDOWS;
 #else
-  rktio->errkind = RKTIO_ERROR_KIND_POSIX;
+  err->errkind = RKTIO_ERROR_KIND_POSIX;
 #endif
 }
-#define set_socket_error(errid) do_set_socket_error(rktio, errid)
+#define set_socket_error(errid) do_set_socket_error(&rktio->err, errid)
 
-static void do_set_gai_error(rktio_t *rktio, int errid) {
-  rktio->errid = errid;
-  rktio->errkind = RKTIO_ERROR_KIND_GAI;
+static void do_set_gai_error(rktio_err_t *err, int errid) {
+  err->errid = errid;
+  err->errkind = RKTIO_ERROR_KIND_GAI;
 }
-#define set_gai_error(err) do_set_gai_error(rktio, err)
+#define set_gai_error(errv) do_set_gai_error(&rktio->err, errv)
 
 #define TCP_BUFFER_SIZE 4096
 
@@ -290,7 +291,7 @@ struct rktio_addrinfo_lookup_t {
 # ifdef RKTIO_SYSTEM_WINDOWS
   HANDLE done_sema;
 # else
-  int done_fd[2];
+  intptr_t done_fd[2];
 # endif
 
   /* For chaining active requests: */
@@ -516,8 +517,10 @@ static rktio_addrinfo_lookup_t *start_lookup(rktio_t *rktio, rktio_addrinfo_look
   
   if (!rktio->ghbn_started) {
     rktio->ghbn_run = 1;
-    if (!ghbn_init(rktio))
+    if (!ghbn_init(rktio)) {
+      free_lookup(lookup);
       return NULL;
+    }
     rktio->ghbn_started = 1;
   }
 
@@ -531,7 +534,7 @@ static rktio_addrinfo_lookup_t *start_lookup(rktio_t *rktio, rktio_addrinfo_look
     }
   }
 # else
-  if (pipe(lookup->done_fd)) {
+  if (rktio_make_os_pipe(rktio, lookup->done_fd, RKTIO_NO_INHERIT_INPUT | RKTIO_NO_INHERIT_OUTPUT)) {
     get_posix_error();
     free_lookup(lookup);
     return NULL;
@@ -721,6 +724,15 @@ rktio_addrinfo_lookup_t *rktio_start_addrinfo_lookup(rktio_t *rktio,
   if (passive) {
     RKTIO_AS_ADDRINFO(hints)->ai_flags |= rktio_AI_PASSIVE;
   }
+  /* On modern systems, `AI_ADDRCONFIG` and `AI_V4MAPPED` tend to be
+     the defaults when `hints` is NULL, so add them if they seem
+     available. */
+#if defined(AI_ADDRCONFIG)
+  RKTIO_AS_ADDRINFO(hints)->ai_flags |= AI_ADDRCONFIG;
+#endif
+#if defined(AI_V4MAPPED) && !defined(__ANDROID__)
+  RKTIO_AS_ADDRINFO(hints)->ai_flags |= AI_V4MAPPED;
+#endif
   if (tcp) {
     RKTIO_AS_ADDRINFO(hints)->ai_socktype = SOCK_STREAM;
 # ifndef PROTOENT_IS_INT
@@ -908,7 +920,7 @@ void rktio_socket_init(rktio_t *rktio, rktio_fd_t *rfd)
 # ifdef SO_BROADCAST
     {
       int bc = 1;
-      setsockopt(s, SOL_SOCKET, SO_BROADCAST, &bc, sizeof(bc));
+      setsockopt(s, SOL_SOCKET, SO_BROADCAST, (char *)&bc, sizeof(bc));
     }
 # endif
 #endif
@@ -967,7 +979,33 @@ int rktio_socket_shutdown(rktio_t *rktio, rktio_fd_t *rfd, int mode)
   return 1;
 }
 
-int rktio_socket_poll_write_ready(rktio_t *rktio, rktio_fd_t *rfd)
+int rktio_tcp_nodelay(rktio_t *rktio, rktio_fd_t *rfd, rktio_bool_t enable)
+{
+  rktio_socket_t s = rktio_fd_socket(rktio, rfd);
+  int nd = (enable ? 1 : 0), r;
+  r = setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (char *)&nd, sizeof(nd));
+  if (r) {
+    get_socket_error();
+    return 0;
+  }
+  return 1;
+}
+
+int rktio_tcp_keepalive(rktio_t *rktio, rktio_fd_t *rfd, rktio_bool_t enable)
+{
+#ifdef SO_KEEPALIVE
+  rktio_socket_t s = rktio_fd_socket(rktio, rfd);
+  int nd = (enable ? 1 : 0), r;
+  r = setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, (char *)&nd, sizeof(nd));
+  if (r) {
+    get_socket_error();
+    return 0;
+  }
+#endif
+  return 1;
+}
+
+int rktio_socket_poll_write_ready(rktio_t *rktio, rktio_fd_t *rfd, rktio_err_t *err)
 {
 #ifdef RKTIO_SYSTEM_UNIX
   return rktio_poll_write_ready(rktio, rfd);
@@ -991,7 +1029,7 @@ int rktio_socket_poll_write_ready(rktio_t *rktio, rktio_fd_t *rfd)
     sr = select(RKTIO_SOCKS(s + 1), NULL, writefds, exnfds, &time);
 
     if (sr == -1) {
-      get_socket_error();
+      do_get_socket_error(err);
       return RKTIO_POLL_ERROR;
     } else if (sr)
       return RKTIO_POLL_READY;
@@ -1001,7 +1039,7 @@ int rktio_socket_poll_write_ready(rktio_t *rktio, rktio_fd_t *rfd)
 #endif
 }
 
-int rktio_socket_poll_read_ready(rktio_t *rktio, rktio_fd_t *rfd)
+int rktio_socket_poll_read_ready(rktio_t *rktio, rktio_fd_t *rfd, rktio_err_t *err)
 {
 #ifdef RKTIO_SYSTEM_UNIX
   return rktio_poll_read_ready(rktio, rfd);
@@ -1025,7 +1063,7 @@ int rktio_socket_poll_read_ready(rktio_t *rktio, rktio_fd_t *rfd)
     sr = select(RKTIO_SOCKS(s + 1), readfds, NULL, exnfds, &time);
     
     if (sr == -1) {
-      get_socket_error();
+      do_get_socket_error(err);
       return RKTIO_POLL_ERROR;
     } else if (sr)
       return RKTIO_POLL_READY;
@@ -1059,11 +1097,13 @@ rktio_fd_t *rktio_socket_dup(rktio_t *rktio, rktio_fd_t *rfd)
 #endif
 }
 
-intptr_t rktio_socket_read(rktio_t *rktio, rktio_fd_t *rfd, char *buffer, intptr_t len)
+intptr_t rktio_socket_read(rktio_t *rktio, rktio_fd_t *rfd, char *buffer, intptr_t len, rktio_err_t *err)
 {
   rktio_socket_t s = rktio_fd_socket(rktio, rfd);
   int rn;
-  
+
+  len = LIMIT_REQUEST_SIZE(len);
+
   do {
     rn = recv(s, buffer, len, 0);
   } while ((rn == -1) && NOT_WINSOCK(errno == EINTR));
@@ -1073,19 +1113,22 @@ intptr_t rktio_socket_read(rktio_t *rktio, rktio_fd_t *rfd, char *buffer, intptr
   else if (rn == 0)
     return RKTIO_READ_EOF;
   else {
-    int err = SOCK_ERRNO();
-    if (WAS_EAGAIN(err))
+    int errv = SOCK_ERRNO();
+    if (WAS_EAGAIN(errv))
       return 0;
     else {
-      get_socket_error();
+      do_get_socket_error(err);
       return RKTIO_READ_ERROR;
     }
   }
 }
 
 static intptr_t do_socket_write(rktio_t *rktio, rktio_fd_t *rfd, const char *buffer, intptr_t len,
+                                rktio_err_t *err,
                                 /* for UDP sendto: */
-                                rktio_addrinfo_t *addr)
+                                rktio_addrinfo_t *addr,
+                                /* alternative address mode for UDP sendto: */
+                                const char *addr_bytes, intptr_t addr_bytes_len)
 {
   rktio_socket_t s = rktio_fd_socket(rktio, rfd);
   intptr_t sent;
@@ -1115,6 +1158,12 @@ static intptr_t do_socket_write(rktio_t *rktio, rktio_fd_t *rfd, const char *buf
         if (!WAS_EBADADDRESS(errid))
           break;
       }
+    } else if (addr_bytes) {
+      do {
+        sent = sendto(s, buffer, len, 0, (const struct sockaddr *)addr_bytes, addr_bytes_len);
+      } while ((sent == -1) && NOT_WINSOCK(errno == EINTR));
+      if (sent < 0)
+        errid = SOCK_ERRNO();
     } else {
       do {
         sent = send(s, buffer, len, 0);
@@ -1133,15 +1182,15 @@ static intptr_t do_socket_write(rktio_t *rktio, rktio_fd_t *rfd, const char *buf
       /* split the message and try again: */
       len >>= 1;
     } else {
-      get_socket_error();
+      do_get_socket_error(err);
       return RKTIO_WRITE_ERROR;
     }
   }
 }
 
-intptr_t rktio_socket_write(rktio_t *rktio, rktio_fd_t *rfd, const char *buffer, intptr_t len)
+intptr_t rktio_socket_write(rktio_t *rktio, rktio_fd_t *rfd, const char *buffer, intptr_t len, rktio_err_t *err)
 {
-  return do_socket_write(rktio, rfd, buffer, len, NULL);
+  return do_socket_write(rktio, rfd, buffer, LIMIT_REQUEST_SIZE(len), err, NULL, NULL, 0);
 }
 
 /*========================================================================*/
@@ -1178,11 +1227,21 @@ static rktio_connect_t *try_connect(rktio_t *rktio, rktio_connect_t *conn)
 {
   struct rktio_addrinfo_t *addr;
   rktio_socket_t s;
-  
+
+  rktio_cloexec_lock();
+
   addr = conn->addr;
   s = socket(RKTIO_AS_ADDRINFO(addr)->ai_family,
              RKTIO_AS_ADDRINFO(addr)->ai_socktype,
              RKTIO_AS_ADDRINFO(addr)->ai_protocol);
+
+  if (s != INVALID_SOCKET)
+    rktio_fd_cloexec(s);
+  else
+    get_socket_error();
+
+  rktio_cloexec_unlock();
+
   if (s != INVALID_SOCKET) {
     int status, inprogress;
     if (!conn->src
@@ -1223,14 +1282,13 @@ static rktio_connect_t *try_connect(rktio_t *rktio, rktio_connect_t *conn)
     }
   }
   
-  get_socket_error();
   return NULL;
 }
 
 int rktio_poll_connect_ready(rktio_t *rktio, rktio_connect_t *conn)
 {
   if (conn->inprogress)
-    return rktio_socket_poll_write_ready(rktio, conn->trying_fd);
+    return rktio_socket_poll_write_ready(rktio, conn->trying_fd, &rktio->err);
   else
     return RKTIO_POLL_READY;
 }
@@ -1373,12 +1431,19 @@ rktio_listener_t *rktio_listen(rktio_t *rktio, rktio_addrinfo_t *src, int backlo
 	}
 #endif
 
+        rktio_cloexec_lock();
+
 	s = socket(RKTIO_AS_ADDRINFO(addr)->ai_family,
                    RKTIO_AS_ADDRINFO(addr)->ai_socktype,
                    RKTIO_AS_ADDRINFO(addr)->ai_protocol);
+
         if (s != INVALID_SOCKET)
+          rktio_fd_cloexec(s);
+        else
           get_socket_error();
 
+        rktio_cloexec_unlock();
+
 #ifdef RKTIO_TCP_LISTEN_IPV6_ONLY_SOCKOPT
 	if (s == INVALID_SOCKET) {
 	  /* Maybe it failed because IPv6 is not available: */
@@ -1394,7 +1459,7 @@ rktio_listener_t *rktio_listen(rktio_t *rktio, rktio_addrinfo_t *src, int backlo
 	    int ok;
 # ifdef IPV6_V6ONLY
 	    int on = 1;
-	    ok = setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY, &on, sizeof(on));
+	    ok = setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY, (char *)&on, sizeof(on));
 # else
 	    ok = -1;
 # endif
@@ -1626,8 +1691,7 @@ void rktio_poll_add_accept(rktio_t *rktio, rktio_listener_t *listener, rktio_pol
 
 rktio_fd_t *rktio_accept(rktio_t *rktio, rktio_listener_t *listener)
 {
-  int
-    ready_pos;
+  int ready_pos;
   rktio_socket_t s, ls;
   rktio_sockopt_len_t l;
   char tcp_accept_addr[RKTIO_SOCK_NAME_MAX_LEN];
@@ -1642,11 +1706,20 @@ rktio_fd_t *rktio_accept(rktio_t *rktio, rktio_listener_t *listener)
 
   l = sizeof(tcp_accept_addr);
 
+  rktio_cloexec_lock();
+
   do {
     s = accept(ls, (struct sockaddr *)tcp_accept_addr, &l);
   } while ((s == -1) && NOT_WINSOCK(errno == EINTR));
 
-  if (s != INVALID_SOCKET) {    
+  if (s != INVALID_SOCKET)
+    rktio_fd_cloexec(s);
+  else
+    get_socket_error();
+
+  rktio_cloexec_unlock();
+
+  if (s != INVALID_SOCKET) {
 # ifdef RKTIO_SYSTEM_UNIX
     RKTIO_WHEN_SET_SOCKBUF_SIZE(int size = TCP_SOCKSENDBUF_SIZE);
     RKTIO_WHEN_SET_SOCKBUF_SIZE(setsockopt(s, SOL_SOCKET, SO_SNDBUF, (char *)&size, sizeof(int)));
@@ -1655,7 +1728,6 @@ rktio_fd_t *rktio_accept(rktio_t *rktio, rktio_listener_t *listener)
     return rktio_system_fd(rktio, s, (RKTIO_OPEN_SOCKET | RKTIO_OPEN_INIT | RKTIO_OPEN_OWN
 				      | RKTIO_OPEN_READ | RKTIO_OPEN_WRITE)); 
   } else {
-    get_socket_error();
     return NULL;
   }
 }
@@ -1744,6 +1816,8 @@ char **rktio_listener_address(rktio_t *rktio, rktio_listener_t *lnr)
 rktio_fd_t *rktio_udp_open(rktio_t *rktio, rktio_addrinfo_t *addr, int family)
 {
   rktio_socket_t s;
+
+  rktio_cloexec_lock();
   
   if (addr)
     s = socket(RKTIO_AS_ADDRINFO(addr)->ai_family,
@@ -1752,10 +1826,15 @@ rktio_fd_t *rktio_udp_open(rktio_t *rktio, rktio_addrinfo_t *addr, int family)
   else
     s = socket(family, SOCK_DGRAM, 0);
 
-  if (s == INVALID_SOCKET) {
+  if (s != INVALID_SOCKET)
+    rktio_fd_cloexec(s);
+  else
     get_socket_error();
+
+  rktio_cloexec_unlock();
+
+  if (s == INVALID_SOCKET)
     return NULL;
-  }
 
   return rktio_system_fd(rktio, s, RKTIO_OPEN_SOCKET | RKTIO_OPEN_UDP | RKTIO_OPEN_INIT);
 }
@@ -1840,7 +1919,7 @@ int rktio_udp_connect(rktio_t *rktio, rktio_fd_t *rfd, rktio_addrinfo_t *addr)
 
 intptr_t rktio_udp_sendto(rktio_t *rktio, rktio_fd_t *rfd, rktio_addrinfo_t *addr, const char *buffer, intptr_t len)
 {
-  return do_socket_write(rktio, rfd, buffer, len, addr);
+  return do_socket_write(rktio, rfd, buffer, len, &rktio->err, addr, NULL, 0);
 }
 
 intptr_t rktio_udp_sendto_in(rktio_t *rktio, rktio_fd_t *rfd, rktio_addrinfo_t *addr, const char *buffer,
@@ -1849,10 +1928,15 @@ intptr_t rktio_udp_sendto_in(rktio_t *rktio, rktio_fd_t *rfd, rktio_addrinfo_t *
   return rktio_udp_sendto(rktio, rfd, addr, buffer + start, end - start);
 }
 
-rktio_length_and_addrinfo_t *rktio_udp_recvfrom(rktio_t *rktio, rktio_fd_t *rfd, char *buffer, intptr_t len)
+intptr_t rktio_udp_sendto_addr_bytes(rktio_t *rktio, rktio_fd_t *rfd, const char *addr, intptr_t addr_len,
+                                     const char *buffer, intptr_t start, intptr_t end)
+{
+  return do_socket_write(rktio, rfd, buffer + start, end - start, &rktio->err, NULL, addr, addr_len);
+}
+
+static void *do_rktio_udp_recvfrom(rktio_t *rktio, rktio_fd_t *rfd, char *buffer, intptr_t len, int decode_address)
 {
   rktio_socket_t s = rktio_fd_socket(rktio, rfd);
-  rktio_length_and_addrinfo_t *r;
   int rn, errid;
   char src_addr[RKTIO_SOCK_NAME_MAX_LEN];
   rktio_sockopt_len_t asize = sizeof(src_addr);
@@ -1893,11 +1977,33 @@ rktio_length_and_addrinfo_t *rktio_udp_recvfrom(rktio_t *rktio, rktio_fd_t *rfd,
       break;
   }
 
-  r = malloc(sizeof(rktio_length_and_addrinfo_t));
-  r->len = rn;
-  r->address = get_numeric_strings(rktio, src_addr, asize);
+  if (decode_address) {
+    rktio_length_and_addrinfo_t *r;
 
-  return r;
+    r = malloc(sizeof(rktio_length_and_addrinfo_t));
+    r->len = rn;
+    r->address = get_numeric_strings(rktio, src_addr, asize);
+
+    return r;
+  } else {
+    rktio_length_and_addr_bytes_t *r;
+    char *addr_bytes;
+
+    addr_bytes = malloc(asize);
+    memcpy(addr_bytes, src_addr, asize);
+
+    r = malloc(sizeof(rktio_length_and_addr_bytes_t));
+    r->len = rn;
+    r->addr_len = asize;
+    r->addr_bytes = addr_bytes;
+
+    return r;
+  }
+}
+
+rktio_length_and_addrinfo_t *rktio_udp_recvfrom(rktio_t *rktio, rktio_fd_t *rfd, char *buffer, intptr_t len)
+{
+  return do_rktio_udp_recvfrom(rktio, rfd, buffer, len, 1);
 }
 
 rktio_length_and_addrinfo_t *rktio_udp_recvfrom_in(rktio_t *rktio, rktio_fd_t *rfd,
@@ -1906,6 +2012,16 @@ rktio_length_and_addrinfo_t *rktio_udp_recvfrom_in(rktio_t *rktio, rktio_fd_t *r
   return rktio_udp_recvfrom(rktio, rfd, buffer + start, end - start);
 }
 
+rktio_length_and_addr_bytes_t *rktio_udp_recvfrom_addr_bytes(rktio_t *rktio, rktio_fd_t *rfd,
+                                                             char *buffer, intptr_t start, intptr_t end)
+{
+  return do_rktio_udp_recvfrom(rktio, rfd, buffer + start, end - start, 0);
+}
+
+char **rktio_addr_bytes_address(rktio_t *rktio, const char *addr, intptr_t len) {
+  return get_numeric_strings(rktio, (void *)addr, len);
+}
+
 int rktio_udp_set_receive_buffer_size(rktio_t *rktio, rktio_fd_t *rfd, int size)
 {
   rktio_socket_t s = rktio_fd_socket(rktio, rfd);
@@ -1923,7 +2039,7 @@ int rktio_udp_set_receive_buffer_size(rktio_t *rktio, rktio_fd_t *rfd, int size)
 int rktio_udp_get_multicast_loopback(rktio_t *rktio, rktio_fd_t *rfd)
 {
   rktio_socket_t s = rktio_fd_socket(rktio, rfd);
-  u_char loop;
+  unsigned char loop;
   rktio_sockopt_len_t loop_len = sizeof(loop);
   int status;
   
@@ -1939,7 +2055,7 @@ int rktio_udp_get_multicast_loopback(rktio_t *rktio, rktio_fd_t *rfd)
 int rktio_udp_set_multicast_loopback(rktio_t *rktio, rktio_fd_t *rfd, int on)
 {
   rktio_socket_t s = rktio_fd_socket(rktio, rfd);
-  u_char loop = (on ? 1 : 0);
+  unsigned char loop = (on ? 1 : 0);
   rktio_sockopt_len_t loop_len = sizeof(loop);
   int status;
   
@@ -1987,7 +2103,7 @@ int rktio_udp_set_ttl(rktio_t *rktio, rktio_fd_t *rfd, int ttl_val)
 int rktio_udp_get_multicast_ttl(rktio_t *rktio, rktio_fd_t *rfd)
 {
   rktio_socket_t s = rktio_fd_socket(rktio, rfd);
-  u_char ttl;
+  unsigned char ttl;
   rktio_sockopt_len_t ttl_len = sizeof(ttl);
   int status;
   
@@ -2003,7 +2119,7 @@ int rktio_udp_get_multicast_ttl(rktio_t *rktio, rktio_fd_t *rfd)
 int rktio_udp_set_multicast_ttl(rktio_t *rktio, rktio_fd_t *rfd, int ttl_val)
 {
   rktio_socket_t s = rktio_fd_socket(rktio, rfd);
-  u_char ttl = ttl_val;
+  unsigned char ttl = ttl_val;
   rktio_sockopt_len_t ttl_len = sizeof(ttl);
   int status;
   
